
/* 
 * $Id: Parser.java,v 1.14 2003/12/08 09:56:47 koto Exp $
 * 
 * $Log: Parser.java,v $
 * Revision 1.14  2003/12/08 09:56:47  koto
 * (1) Bug Fix about $getChildNodes (2) Remove embed feature.
 *
 * Revision 1.13  2003/12/03 06:50:13  koto
 * (1) Support absorbers  (2) Bug fix about getChildNodes()
 *
 * Revision 1.12  2003/12/02 14:27:20  koto
 * Remove $ambiguous and $absorber.
 *
 * Revision 1.11  2003/09/04 07:55:24  koto
 * Implement typed absorber.  $ambiguous aliases are not implemented yet.
 *
 * Revision 1.10  2003/09/04 05:30:12  koto
 * Make ambiguity be represented by symbol rather than reduction/LRState.
 *
 * Revision 1.9  2003/09/04 03:42:44  koto
 * Disuse List list(List).
 *
 * Revision 1.8  2003/08/20 12:54:54  koto
 * (1) bug fix; IDENTIFIER and RESERVED_WORD (2) Add  (everytime valid).
 *
 * Revision 1.7  2003/08/17 06:53:26  koto
 * BugFix
 *
 * Revision 1.6  2003/08/16 16:38:36  koto
 * Add separated expressions.
 *
 * 
 */

package jp.gr.java_conf.koto.notavacc.parser;

import java.io.*;
import java.util.*;
import jp.gr.java_conf.koto.notavacc.*;
import jp.gr.java_conf.koto.notavacc.util.*;

public class Parser extends Original {
    protected Node createNode(int symbolID, NodeInitializationParameters parameters) throws ParseException {
        Node result;
        switch (symbolID) {
         default:
            result = super.createNode(symbolID, parameters, true);
            break;
        case Root.ID:
            result = new Root(environment, parameters);
            break;
        case InlineExpression.ID:
            result = new InlineExpressionReplacement(parameters);
            break;
        }
        
        // replace "++" expressions and "**" expressions.
        if (result instanceof BinaryTokenExpression) {
            BinaryTokenExpression n = (BinaryTokenExpression) result;
            
            class SequentialTokenExpression extends Default.SequentialTokenExpression {
                SequentialTokenExpression(TokenExpression op1, TokenExpression op2) {
                    super(new LinkedList(), Arrays.asList(new TokenExpression[] { op1, op2 }));
                }
            }
            class OptionalTokenExpression extends Default.SelectiveTokenExpression {
                OptionalTokenExpression(TokenExpression op1) {
                    super(new LinkedList(), Arrays.asList(new TokenExpression[] { op1, new Default.SequentialTokenExpression(new LinkedList(), Collections.EMPTY_LIST) }));
                }
            }
            class PlusTokenExpression extends Default.PlusTokenExpression {
                PlusTokenExpression(TokenExpression op1) {
                    super(new LinkedList(), op1);
                }
            }
            
            final Default.TokenExpression op2op1star = new OptionalTokenExpression(new PlusTokenExpression(new SequentialTokenExpression(n.operand2(), n.operand1())));
            Default.TokenExpression oneOrMore = new SequentialTokenExpression(n.operand1(), op2op1star) {
                {
                    // to avoid visiting n.operand1() twice.
                    getChildNodes().clear();
                    getChildNodes().add(op2op1star);
                }
            };
            
            if (result instanceof OneOrMoreSeparatedTokenExpression) {
                return oneOrMore;
            } else {
                assert result instanceof ZeroOrMoreSeparatedTokenExpression;
                Default.TokenExpression zeroOrMore = new OptionalTokenExpression(oneOrMore);
                return zeroOrMore;
            }
        } else if (result instanceof BinaryExpression) {
            BinaryExpression n = (BinaryExpression) result;
            
            class SequentialExpression extends Default.SequentialExpression {
                SequentialExpression(Expression op1, Expression op2) {
                    super(new LinkedList(), Arrays.asList(new Expression[] { op1, op2 }));
                }
            }
            class OptionalExpression extends Default.SelectiveExpression {
                OptionalExpression(Expression op1) {
                    super(new LinkedList(), Arrays.asList(new Expression[] { op1, new Default.SequentialExpression(new LinkedList(), Collections.EMPTY_LIST) }));
                }
            }
            class PlusExpression extends Default.PlusExpression {
                PlusExpression(Expression op1, Token mark) {
                    super(new LinkedList(), op1, mark);
                }
            }
            
            final Default.Expression op2op1star = new OptionalExpression(new PlusExpression(new SequentialExpression(n.operand2(), n.operand1()), n.mark()));
            Default.Expression oneOrMore = new SequentialExpression(n.operand1(), op2op1star) {
                {
                    // to avoid visiting n.operand1() twice.
                    getChildNodes().clear();
                    getChildNodes().add(op2op1star);
                }
            };
            
            if (result instanceof OneOrMoreSeparatedExpression) {
                return oneOrMore;
            } else {
                assert result instanceof ZeroOrMoreSeparatedExpression;
                Default.Expression zeroOrMore = new OptionalExpression(oneOrMore);
                return zeroOrMore;
            }
        }
        return result;
    }
    
    protected LexicalAnalyzer createLexicalAnalyzer(String sourceName, CharSequence text, int tabStop) throws ParseException {
        return new Default.LexicalAnalyzer(sourceName, text, tabStop) {
            protected char nextChar() throws ParseException {
                return super.nextUnicodeEscapedChar();
            }
            public Token next() throws ParseException {
                Token result = super.next();
                if (result.getSymbolID() == TOKEN_IDENTIFIER) {
                    String image = result.getImage();
                    int i = 0;
                    for (; i < image.length(); i++) {
                        char ch = image.charAt(i);
                        if (i == 0) {
                            if (!Character.isJavaIdentifierStart(ch))
                                break;
                        } else {
                            if (!Character.isJavaIdentifierPart(ch))
                                break;
                        }
                    }
                    if (i < image.length())
                        throw new ParseException("The " + i + "th letter \\u" + Integer.toHexString(image.charAt(i)) +  " of the token " + result + " cannot be used.", result);
                }
                return result;
            }
        };
    }
    
    private Environment environment;
    public Parser(Environment env) {
        this.environment = env;
    }
    
    public static class Root extends Default.Root {
        private Logger logger;
        private Environment environment;
        public Root(Environment env, NodeInitializationParameters parameters) throws ParseException {
            super(parameters, true);
            this.environment = env;
            this.logger = env.logger;
            
            final int[] count = { 0 };
            new Visitor() {
                protected void visit(EmbedExpression exp) {
                    switch (count[0]++) {
                    case 0:
                        logger.warn(exp.mark(), "Now, $embed expressions are not embeded.  This warning is reported only once for a file.");
                        break;
                    }
                }
            }.visitAll(this);
            
            checkErrors();
            buildDefinitionMap();
            bindNames();
            buildSetOfDefinitionsHavingExplicitParameterLabel();
            
            //checkNameUse();   $token x = "str"; で、"str" が使われている場合に警告しないためには、ここではチェックできない。
            
            //formatAbsorbers();
        }
        
        private void checkErrors() throws ParseException {
            /*if (!environment.experimental) {
                new Visitor() {
                    protected void visit(TokenDefinition def) {
                        if (def.weak() != null)
                            logger.error("{0} is not acceptable.", def.weak());
                    }
                }.visitAll(this);
            }*/
        }
        
        public boolean isProtectedParser() {
            return parserDeclaration().protectedKeyword() != null;
        }
        
        private String[] parserName = null;
        private String[] getParserName() {
            if (parserName != null)
                return parserName;
            
            List identifiers = parserDeclaration().name().identifiers();
            parserName = new String[] {
                (identifiers.size() <= 1) ? null : toString(identifiers.subList(0, identifiers.size() - 1)),
                ((Token) identifiers.get(identifiers.size() - 1)).getImage(),
            };
            return parserName;
        }
        private String toString(List identifiers) {
            StringBuffer result = new StringBuffer();
            Iterator it0 = identifiers.iterator();
            while (it0.hasNext()) {
                Token identifier = (Token) it0.next();
                result.append(identifier.getImage());
                if (it0.hasNext())
                    result.append(".");
            }
            return result.toString();
        }
        public String getPackageName() {
            return getParserName()[0];
        }
        public String getParserSimpleName() {
            return getParserName()[1];
        }
        
        public boolean hasProtectedConstructor() {
            return constructorScope() != null;
        }
        
        private List typeDefinitions = null;
        public List typeDefinitions() {
            if (typeDefinitions == null) {
                typeDefinitions = new LinkedList();
                new Visitor() {
                    protected void visit(TypeDefinition def) {
                        typeDefinitions.add(def);
                    }
                }.visitAll(this);
            }
            return typeDefinitions;
        }
        
        public boolean supportsAbsorber() {
            return !absorberDefinitions2().isEmpty();
        }
        
        private Map aliasDefinitionToAbsorbCandidateFlag = new HashMap();
        public boolean isAbsorbCandidate(AliasDefinition definition) {
            Boolean result = (Boolean) aliasDefinitionToAbsorbCandidateFlag.get(definition);
            if (result == null) {
                final int SHORTER = 0;
                final int LONGER = 1;
                Map nameToRange = new LinkedHashMap();
                new ExpressionVisitor(this, 2) {
                    protected void visit(Context context, SelectiveExpression expr, Object argument) {
                        Map nameToRange = (Map) argument;
                        
                        Set names = new LinkedHashSet();
                        List subMaps = new LinkedList();
                        
                        Iterator it32 = expr.operands().iterator();
                        while (it32.hasNext()) {
                            Expression operand = (Expression) it32.next();
                            
                            Map sub = new LinkedHashMap();
                            visitAll(context, operand, sub);
                            
                            names.addAll(sub.keySet());
                            subMaps.add(sub);
                        }
                        
                        Iterator it6 = names.iterator();
                        while (it6.hasNext()) {
                            String name = (String) it6.next();
                            
                            int[] range = new int[2];
                            range[SHORTER] = Integer.MAX_VALUE;
                            range[LONGER] = 0;
                            
                            Iterator it7 = subMaps.iterator();
                            while (it7.hasNext()) {
                                Map subMap = (Map) it7.next();
                                int[] subRange = (int[]) subMap.get(name);
                                if (subRange == null) {
                                    range[SHORTER] = Math.min(range[SHORTER], 0);
                                    range[LONGER] = Math.max(range[LONGER], 0);
                                } else {
                                    range[SHORTER] = Math.min(range[SHORTER], subRange[SHORTER]);
                                    range[LONGER] = Math.max(range[LONGER], subRange[LONGER]);
                                }
                            }
                            
                            nameToRange.put(name, range);
                        }
                    }
                    protected void visit(Context context, SequentialExpression expr, Object argument) {
                        Map nameToRange = (Map) argument;
                        
                        Iterator it32 = expr.operands().iterator();
                        while (it32.hasNext()) {
                            Expression operand = (Expression) it32.next();
                            
                            Map sub = new LinkedHashMap();
                            visitAll(context, operand, sub);
                            
                            Iterator it5 = sub.entrySet().iterator();
                            while (it5.hasNext()) {
                                Map.Entry e = (Map.Entry) it5.next();
                                String name = (String) e.getKey();
                                int[] subRange = (int[]) e.getValue();
                                
                                int[] range = (int[]) nameToRange.get(name);
                                if (range == null) {
                                    nameToRange.put(name, subRange);
                                } else {
                                    range[SHORTER] += subRange[SHORTER];
                                    range[LONGER] += subRange[LONGER];
                                }
                            }
                        }
                    }
                    protected void visit(Context context, PlusExpression expr, Object argument) {
                        Map nameToRange = (Map) argument;
                        
                        super.visit(context, expr, argument);
                        
                        Iterator it8 = nameToRange.values().iterator();
                        while (it8.hasNext()) {
                            int[] range = (int[]) it8.next();
                            range[LONGER] *= 2;
                        }
                    }
                    protected void visit(Context context, LabeledExpression expr, Object argument) {
                        visitAll(context, expr.operand(), argument);
                    }
                    protected void visit(Context context, IdentifierExpression expr, AliasDefinition def, Object argument) {
                        super.visit(context, expr, def, argument);
                    }
                    protected void visit(Context context, IdentifierExpression expr, TypeDefinition def, Object argument) {
                        countUp(context, argument);
                    }
                    protected void visit(Context context, IdentifierExpression expr, TokenReservation def, Object argument) {
                        countUp(context, argument);
                    }
                    protected void visit(Context context, StringExpression expr, Object argument) {
                        countUp(context, argument);
                    }
                    private void countUp(Context context, Object argument) {
                        Map nameToRange = (Map) argument;
                        
                        Iterator it4 = context.labelStrings.iterator();
                        while (it4.hasNext()) {
                            String label = (String) it4.next();
                            int[] range = (int[]) nameToRange.get(label);
                            if (range == null) {
                                range = new int[2];
                                nameToRange.put(label, range);
                            }
                            
                            range[SHORTER] = 1;
                            range[LONGER] = 1;
                        }
                        
                        nameToRange.put(null, new int[] { 1, 1 });
                    }
                }.visitAll(definition.expression(), nameToRange);
                
                Iterator it3 = nameToRange.values().iterator();
                while (it3.hasNext()) {
                    int[] range = (int[]) it3.next();
                    if (range[SHORTER] != 1 || range[LONGER] != 1) {
                        aliasDefinitionToAbsorbCandidateFlag.put(definition, Boolean.FALSE);
                        return false;
                    }
                }
                result = Boolean.valueOf(!nameToRange.isEmpty());
                aliasDefinitionToAbsorbCandidateFlag.put(definition, result);
            }
            return result.booleanValue();
        }
        
        public String getCanonicalName(Definition definition) {
            return definition.identifier().getImage();
        }
        
        private Map canonicalNameToDefinition;
        private void buildDefinitionMap() {
            canonicalNameToDefinition = new LinkedHashMap();
            final Map duplexNames = new LinkedHashMap();
            new Visitor() {
                protected void visit(Definition def) {
                    if (def.identifier() != null) {
                        String canonicalName = getCanonicalName(def);
                        Definition olddef = (Definition) canonicalNameToDefinition.put(canonicalName, def);
                        if (olddef != null) {
                            List definitions = (List) duplexNames.get(canonicalName);
                            if (definitions == null) {
                                definitions = new LinkedList();
                                duplexNames.put(canonicalName, definitions);
                                definitions.add(olddef);
                            }
                            definitions.add(def);
                        }
                    }
                }
            }.visitAll(this);
            if (!duplexNames.isEmpty()) {
                Iterator it = duplexNames.values().iterator();
                while (it.hasNext()) {
                    List list = (List) it.next();
                    Iterator it2 = list.iterator();
                    while (it2.hasNext()) {
                        Definition def = (Definition) it2.next();
                        logger.error("{0} defined twice or more.", def.identifier());
                    }
                }
            }
            if (logger.hasError())
                logger.fatal();
        }
        private Map nameToDefinition;
        private void bindNames() {
            nameToDefinition = new LinkedHashMap();
            new Visitor() {
                protected void visit(Name name) {
                    String canonicalName = name.identifier().getImage();
                    Definition def = (Definition) canonicalNameToDefinition.get(canonicalName);
                    nameToDefinition.put(name, def);
                    
                    if (def == null) {
                        logger.error("{0} is not defined.", name.identifier());
                    } else if (name instanceof TypeName) {
                        if (!(def instanceof TypeDefinition))
                            logger.error("{0} should be a type name.", name.identifier());
                    } else if (name instanceof SubtokenName) {
                        if (!(def instanceof SubtokenDefinition))
                            logger.error("{0} should be a token name or a subtoken name.", name.identifier());
                        else if ((def instanceof TokenDefinition) && ((TokenDefinition) def).weak() != null)
                            logger.error("cannot refer $weak {0}.", name.identifier());
                    } else if (name instanceof SymbolName) {
                        boolean ok = false;
                        if (def instanceof TokenReservation) {
                            if (!(def instanceof WhiteTokenDefinition))
                                ok = true;
                        } else if (def instanceof TypeDefinition) {
                            ok = true;
                        } else if (def instanceof AliasDefinition) {
                            ok = true;
                        }
                        if (!ok)
                            logger.error("{0} should be a token/type/alias name.", name.identifier());
                    } else {
                        assert name instanceof AbsorberOrTypeName;
                        if (!(def instanceof AbsorberDefinition2 || def instanceof TypeDefinition))
                            logger.error("{0} should be an absorber name.", name.identifier());
                    }
                }
            }.visitAll(this);
            if (logger.hasError())
                logger.fatal();
        }
        public Definition getDefinition(Name name) {
            Definition result = (Definition) nameToDefinition.get(name);
            assert result != null : "A definition missing for " + name.identifier();
            return result;
        }
        public TypeDefinition getDefinition(TypeName name) {
            return (TypeDefinition) getDefinition((Name) name);
        }
        public SubtokenDefinition getDefinition(SubtokenName name) {
            return (SubtokenDefinition) getDefinition((Name) name);
        }
        public TypeGenerateDefinition getDefinition(AbsorberOrTypeName name) {
            return (TypeGenerateDefinition) getDefinition((Name) name);
        }
        
        
        private Set definitionsHavingExplicitParameterLabel;
        private void buildSetOfDefinitionsHavingExplicitParameterLabel() {
            definitionsHavingExplicitParameterLabel = new LinkedHashSet();
            Iterator it = aliasDefinitions().iterator();
            while (it.hasNext()) {
                final AliasDefinition def = (AliasDefinition) it.next();
                new Visitor() {
                    protected void visit(LabeledExpression exp) {
                        if (exp.label().getImage().equals("$label"))
                            definitionsHavingExplicitParameterLabel.add(def);
                    }
                }.visitAll(def);
            }
        }
        public boolean hasExplicitParameterLabel(AliasDefinition definition) {
            return definitionsHavingExplicitParameterLabel.contains(definition);
        }
        
        /*private void checkNameUse() {
            final Set definitions = new LinkedHashSet();
            new Visitor() {
                protected void visit(Definition def) {
                    definitions.add(def);
                }
            }.visitAll(this);
            final Set used = new LinkedHashSet();
            
            Iterator it1 = definitions.iterator();
            while (it1.hasNext()) {
                Definition def = (Definition) it1.next();
                if (def instanceof TypeDefinition) {
                    TypeDefinition td = (TypeDefinition) def;
                    if (td.parsableKeyword() != null)
                        used.add(definitions);
                } else if (def instanceof WhiteTokenReservation) {
                    // including WhiteTokenDefinition.
                    used.add(definitions);
                } else if (def instanceof TokenReservation && !(def instanceof TokenDefinition)) {
                    used.add(definitions);
                }
            }
            
        }*/
        
        
        /*private void formatAbsorbers() {
            rejectDeadSymbols();
            if (logger.hasError())
                return;
            
            
        }
        private void rejectDeadSymbols() {
            Set unfixedDefinitions = new LinkedHashSet();
            unfixedDefinitions.addAll(typeDefinitions());
            unfixedDefinitions.addAll(aliasDefinitions());
            
            Set livingDefinitions = new LinkedHashSet();
            
            for (;;) {
                boolean active = false;
                
                Iterator it = unfixedDefinitions.iterator();
                while (it.hasNext()) {
                    Definition definition = (Definition) it.next();
                    if (isLiving(definition, livingDefinitions)) {
                        livingDefinitions.add(definition);
                        it.remove();
                        active = true;
                    }
                }
                
                if (!active)
                    break;
            }
            
            if (!unfixedDefinitions.isEmpty()) {
                Iterator it91 = unfixedDefinitions.iterator();
                while (it91.hasNext()) {
                    Definition definition = (Definition) it91.next();
                    logger.error("{0} is a dead symbol; it can match only infinte sequences.", definition.identifier());
                }
            }
        }
        private boolean isLiving(Definition definition, final Set livingDefinitions) {
            Expression expr;
            if (definition instanceof TypeDefinition) {
                expr = ((TypeDefinition) definition).expression();
            } else {
                expr = ((AliasDefinition) definition).expression();
            }
            
            boolean[] result = { false };
            new ExpressionVisitor(this) {
                protected void visit(Context context, SelectiveExpression expr, Object argument) {
                    Iterator it = expr.operands().iterator();
                    while (it.hasNext()) {
                        Expression operand = (Expression) it.next();
                        boolean[] living = { false };
                        visitAll(context, operand, living);
                        if (living[0]) {
                            ((boolean[]) argument)[0] = true;
                            return;
                        }
                    }
                    ((boolean[]) argument)[0] = false;
                }
                protected void visit(Context context, SequentialExpression expr, Object argument) {
                    Iterator it = expr.operands().iterator();
                    while (it.hasNext()) {
                        Expression operand = (Expression) it.next();
                        boolean[] living = { false };
                        visitAll(context, operand, living);
                        if (!living[0]) {
                            ((boolean[]) argument)[0] = false;
                            return;
                        }
                    }
                    ((boolean[]) argument)[0] = true;
                }
                protected void visit(Context context, IdentifierExpression expr, TypeDefinition def, Object argument) {
                    ((boolean[]) argument)[0] = livingDefinitions.contains(def);
                }
                protected void visit(Context context, IdentifierExpression expr, AliasDefinition def, Object argument) {
                    ((boolean[]) argument)[0] = livingDefinitions.contains(def);
                }
                protected void visit(Context context, IdentifierExpression expr, TokenReservation def, Object argument) {
                    ((boolean[]) argument)[0] = true;
                }
                protected void visit(Context context, StringExpression expr, Object argument) {
                    ((boolean[]) argument)[0] = true;
                }
            }.visitAll(expr, result);
            return result[0];
        }*/
        
        public String getValue(StringLiteral literal) {
            StringBuffer result = new StringBuffer();
            String image = literal.token().getImage();
            int maz = image.length() - 1;
            for (int i = 1; i < maz;) {
                char ch = image.charAt(i++);
                if (ch != '\\') {
                    result.append(ch);
                } else {
                    assert i < maz;
                    ch = image.charAt(i++);
                    switch(ch) {
                    case 'n':   result.append('\n');    break;
                    case 't':   result.append('\t');    break;
                    case 'b':   result.append('\b');    break;
                    case 'r':   result.append('\r');    break;
                    case 'f':   result.append('\f');    break;
                    
                    case '\\':
                    case '\'':
                    case '\"':
                        result.append(ch);
                        break;
                    
                    default:
                        assert Character.digit(ch, 8) != -1;
                        {
                            StringBuffer octal = new StringBuffer();
                            octal.append(ch);
                            int max = ('0' <= ch && ch <= '3') ? 3 : 2;
                            while (--max > 0) {
                                if (i >= maz)
                                    break;
                                ch = image.charAt(i);
                                if (Character.digit(ch, 8) == -1)
                                    break;
                                octal.append(ch);
                                i++;
                            }
                            result.append((char) Integer.parseInt(octal.toString(), 8));
                        }
                        break;
                    }
                }
            }
            return result.toString();
        }
        
        public char getValue(CharacterLiteral literal) {
            String image = literal.token().getImage();
            char ch = image.charAt(1);
            if (ch != '\\') {
                return ch;
            }
            
            ch = image.charAt(2);
            switch(ch) {
            case 'n':   return '\n';
            case 't':   return '\t';
            case 'b':   return '\b';
            case 'r':   return '\r';
            case 'f':   return '\f';
            
            case '\\':
            case '\'':
            case '\"':
                return ch;
            
            default:
                assert Character.digit(ch, 8) != -1;
                String octal = image.substring(2, image.length() - 1);
                return (char) Integer.parseInt(octal, 8);
            }
        }
    }
    
    private class InlineExpressionReplacement extends Default.InlineExpression {
        public InlineExpressionReplacement(NodeInitializationParameters parameters) {
            super(parameters, true);
            name = new SymbolNameReplacement();
            getChildNodes().add(0, name);
        }
        
        private final SymbolName name;
        public SymbolName name() {
            return name;
        }
        
        private class SymbolNameReplacement extends Default.Node implements SymbolName {
            public SymbolNameReplacement() {
                super(Collections.singletonList(typeDefinition().identifier()));
            }
            public Original.Token identifier() {
                return typeDefinition().identifier();
            }
        }
    }
    
}

